/*  Lziprecover - Data recovery tool for the lzip format
    Copyright (C) 2009-2018 Antonio Diaz Diaz.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _FILE_OFFSET_BITS 64

#include <cerrno>
#include <cstdio>
#include <cstring>
#include <string>
#include <vector>
#include <stdint.h>
#include <unistd.h>
#include <utime.h>
#include <sys/stat.h>

#include "lzip.h"
#include "block.h"
#include "lzip_index.h"


int dump_damaged( const std::vector< std::string > & filenames,
                  const std::string & default_output_filename,
                  const bool force, const bool strip,
                  const bool ignore_trailing, const bool loose_trailing,
                  const bool strip_trailing )
  {
  if( default_output_filename.empty() ) outfd = STDOUT_FILENO;
  else
    {
    output_filename = default_output_filename;
    if( !open_outstream( force, true, false, false ) ) return 1;
    }
  unsigned long long total_size = 0;
  int files = 0, members = 0, retval = 0;
  bool stdin_used = false;
  for( unsigned i = 0; i < filenames.size(); ++i )
    {
    const bool from_stdin = ( filenames[i] == "-" );
    if( from_stdin ) { if( stdin_used ) continue; else stdin_used = true; }
    const char * const input_filename =
      from_stdin ? "(stdin)" : filenames[i].c_str();
    struct stat in_stats;				// not used
    const int infd = from_stdin ? STDIN_FILENO :
      open_instream( input_filename, &in_stats, true, true );
    if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }

    const Lzip_index lzip_index( infd, ignore_trailing, loose_trailing,
                                 true, true );
    if( lzip_index.retval() != 0 )
      {
      show_file_error( input_filename, lzip_index.error().c_str() );
      cleanup_and_fail( lzip_index.retval() );
      }
    if( !safe_seek( infd, 0 ) ) cleanup_and_fail( 1 );
    long long stream_pos = 0;		// first pos not yet read from file
    const int prev_members = members;
    for( long j = 0; j < lzip_index.members(); ++j )
      {
      const Block & mb = lzip_index.mblock( j );
      if( !strip && mb.pos() > stream_pos )			// gap
        {
        if( !safe_seek( infd, stream_pos ) ||
            !copy_file( infd, outfd, mb.pos() - stream_pos ) )
          cleanup_and_fail( 1 );
        total_size += mb.pos() - stream_pos; ++members;
        }
      if( !safe_seek( infd, mb.pos() ) ) cleanup_and_fail( 1 );	// member
      const bool good = ( test_member_from_file( infd, mb.size() ) == 0 );
      if( strip == good )
        {
        if( !safe_seek( infd, mb.pos() ) ||
            !copy_file( infd, outfd, mb.size() ) )
          cleanup_and_fail( 1 );
        total_size += mb.size(); ++members;
        }
      stream_pos = mb.end();
      }
    if( strip && members == prev_members )	// all members were stripped
      { show_file_error( input_filename, "All members are damaged, skipping." );
        close( infd ); continue; }
    const unsigned long long cdata_size = lzip_index.cdata_size();
    const long long trailing_size = lzip_index.file_size() - cdata_size;
    if( strip && !strip_trailing && trailing_size > 0 )	// copy trailing data
      {
      if( !safe_seek( infd, cdata_size ) ||
          !copy_file( infd, outfd, trailing_size ) )
        cleanup_and_fail( 1 );
      if( i + 1 < filenames.size() )
        { show_file_error( input_filename, "Trailing data found. Stopped." );
          retval = 2; break; }
      }
    close( infd );
    }
  if( close_outstream( 0 ) != 0 && retval < 1 ) retval = 1;
  if( verbosity >= 1 )
    {
    if( strip )
      std::fprintf( stderr, "%llu bytes copied from %d member(s) from %d file(s).\n",
                    total_size, members, files );
    else
      std::fprintf( stderr, "%llu damaged bytes dumped from %d member(s) from %d file(s).\n",
                    total_size, members, files );
    }
  return retval;
  }


int remove_damaged( const std::vector< std::string > & filenames,
                    const bool ignore_trailing, const bool loose_trailing,
                    const bool remove_trailing )
  {
  unsigned long long total_size = 0;
  int files = 0, members = 0, retval = 0;
  for( unsigned i = 0; i < filenames.size(); ++i )
    {
    const char * const filename = filenames[i].c_str();
    struct stat in_stats, dummy_stats;
    const int infd = open_instream( filename, &in_stats, true, true );
    if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }

    const Lzip_index lzip_index( infd, ignore_trailing, loose_trailing,
                                 true, true );
    if( lzip_index.retval() != 0 )
      {
      show_file_error( filename, lzip_index.error().c_str() );
      if( retval < lzip_index.retval() ) retval = lzip_index.retval();
      close( infd );
      continue;
      }
    const int fd = open_truncable_stream( filename, &dummy_stats );
    if( fd < 0 ) { close( infd ); if( retval < 1 ) retval = 1; continue; }

    if( !safe_seek( infd, 0 ) ) return 1;
    long long stream_pos = 0;		// first pos not yet written to file
    bool error = false;
    for( long i = 0; i < lzip_index.members(); ++i )
      {
      const Block & mb = lzip_index.mblock( i );
      if( !safe_seek( infd, mb.pos() ) )
        { error = true; if( retval < 1 ) retval = 1; break; }
      if( test_member_from_file( infd, mb.size() ) == 0 )
        {
        if( stream_pos != mb.pos() &&
            ( !safe_seek( infd, mb.pos() ) ||
              !safe_seek( fd, stream_pos ) ||
              !copy_file( infd, fd, mb.size() ) ) )
          { error = true; if( retval < 1 ) retval = 1; break; }
        stream_pos += mb.size();
        }
      else ++members;
      }
    if( error ) { close( fd ); close( infd ); break; }
    if( stream_pos == 0 )			// all members were removed
      { show_file_error( filename, "All members are damaged, skipping." );
        close( fd ); close( infd ); continue; }
    const long long cdata_size = lzip_index.cdata_size();
    if( stream_pos == cdata_size )		// no members were removed
      { close( fd ); close( infd ); continue; }
    total_size += cdata_size - stream_pos; ++files;
    const long long trailing_size = lzip_index.file_size() - cdata_size;
    if( !remove_trailing && trailing_size > 0 )	// copy trailing data
      {
      if( !safe_seek( infd, cdata_size ) ||
          !safe_seek( fd, stream_pos ) ||
          !copy_file( infd, fd, trailing_size ) )
        { close( fd ); close( infd ); if( retval < 1 ) retval = 1; break; }
      stream_pos += trailing_size;
      }
    int result;
    do result = ftruncate( fd, stream_pos );
      while( result != 0 && errno == EINTR );
    if( result != 0 )
      {
      show_file_error( filename, "Can't truncate file", errno );
      if( retval < 1 ) { retval = 1; } break;
      }
    if( close( fd ) != 0 || close( infd ) != 0 )
      {
      show_file_error( filename, "Error closing file", errno );
      if( retval < 1 ) { retval = 1; } break;
      }
    struct utimbuf t;
    t.actime = in_stats.st_atime;
    t.modtime = in_stats.st_mtime;
    utime( filename, &t );
    }
  if( verbosity >= 1 )
    std::fprintf( stderr, "%llu damaged bytes removed from %d member(s) from %d file(s).\n",
                  total_size, members, files );
  return retval;
  }
